from .__base import FarasaBase


class FarasaSegmenter(FarasaBase):
    task = "segment"

    def segment(self, text):
        return self._do_task(text=text)

    def _desegment_word(self, word: str) -> str:
        desegmented_word = word.replace("ل+ال+", "لل")
        if "ال+ال" not in word:
            desegmented_word = desegmented_word.replace("ل+ال", "لل")
        desegmented_word = desegmented_word.replace("+", "")
        desegmented_word = desegmented_word.replace("للل", "لل")
        return desegmented_word

    def desegment(self, text, separator=" "):
        return " ".join(self._desegment_word(word) for word in text.split(separator))



from .__base import FarasaBase


class TaggedToken:
    def __init__(self, token, tag):
        self.tokens, self.tags = list(), list()
        self.tokens.append(token)
        self.tags.append(tag)

    def append_connected_token(self, token=None, tag=None, tagged_token=None):
        assert not (token and tag) or not tagged_token
        if tagged_token:
            self.tokens.extend(tagged_token.tokens)
            self.tags.extend(tagged_token.tags)

        else:
            self.tokens.append(token)
            self.tags.append(tag)

    @property
    def last_token(self):
        return self.tokens[-1]

    # def raise_value_error(self):
    #     raise ValueError("Tagged Token object cannot have more than three subtokens")

    def _process_tag(self, tag):
        subtags = tag.split("-")
        if len(subtags) == 1:
            return subtags
        else:
            return (subtags[0], *list(subtags[1]))

    def process_tags(self):
        processed_tags = list()
        for tag in self.tags:
            processed_subtags = list()
            subtags = tag.split("-")
            processed_subtags.append(subtags[0])
            if len(subtags) > 1:
                for subtag in subtags[1:]:
                    processed_subtags.append(tuple(subtag))
            processed_tags.append(processed_subtags)
        return tuple(processed_tags)

    def as_tuple(self):
        return (
            "".join(token.strip() for token in self.tokens).replace("+", ""),
            # *list(self._process_tag(tag) for tag in self.tags),
            self.process_tags(),
        )

    def __str__(self):
        return str(self.as_tuple())

    def __repr__(self):
        return str(self.as_tuple())


class FarasaPOSTagger(FarasaBase):
    task = "POS"

    def tag(self, text):
        return self._do_task(text=text)

    def tag_segments(self, text, combine_subtokens=False):
        tokens_objects = list()
        tagged_text = self.tag(text)
        # ignore the first and last as they are indications of
        #  the begining and end of the string
        tagged_tokens = tagged_text.split()[1:-1]
        if not combine_subtokens:
            for tagged_token in tagged_tokens:
                token, tag = tagged_token.split("/")
                token_object = TaggedToken(token, tag)
                if not tokens_objects:
                    tokens_objects.append(token_object)
                elif tokens_objects[-1].last_token[-1] == "+":
                    tokens_objects[-1].append_connected_token(tagged_token=token_object)
                elif token_object.last_token[0] == "+":
                    tokens_objects[-1].append_connected_token(tagged_token=token_object)
                else:
                    tokens_objects.append(token_object)
            return tokens_objects
        else:
            tokens_objects = [
                TaggedToken(*tagged_token.split("/")) for tagged_token in tagged_tokens
            ]
            return tokens_objects




import os
import subprocess
import shlex
import warnings
import time
import re
import tempfile

class Farasa():
    temp_filename_in = f'.farasa.{time.time()}.in.temp'
    temp_filename_out = f'.farasa.{time.time()}.out.temp'
    interactive = None # either None or 'interactive'
    APIs_map = {
        'segment':shlex.split('java -jar ./farasa/FarasaSegmenterJar/FarasaSegmenterJar.jar'),
        'stem' : shlex.split('java -jar ./farasa/FarasaSegmenterJar/FarasaSegmenterJar.jar -l true'),
    }
    tasks = list(APIs_map.keys())
    task_proc = None

    def __init__(self, interactive_task=None):
        print("perform system check...")
        self._check_java_version()
        if interactive_task:
            if interactive_task not in self.tasks:
                raise Exception(f"task can only be one of [{', '.join(t for t in self.tasks)}]")
            self.task = interactive_task
            print(f"initialize {self.task} in interactive mode...")
            self.task_proc = subprocess.Popen(self.APIs_map[self.task],
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
            


    def _check_java_version(self):
        try:
            version_proc = subprocess.check_output(['java', '-version'], stderr=subprocess.STDOUT,encoding='utf8',text=True)
            version_pattern = r'\"(\d+\.\d+).*\"'
            java_version = float(re.search(version_pattern,version_proc).groups()[0])
            if java_version >= 1.7:
                print(f"Your java version is {java_version} which is compatiple with Farasa ")
            else:
                warnings.warn('You are using old version of java. Farasa is compatiple with Java 7 and above ')
        except subprocess.CalledProcessError as proc_err:
            print(proc_err)
            raise Exception("We could not check for java version on the machine. Please make sure you have installed Java 1.7+ and add it to your PATH.")

    
    def _write_to_temp_file(self, text):
        assert text
        with open(self.temp_filename_in,'w',encoding='utf-8') as tmp:
            tmp.write(text)


    def _read_from_temp_file(self):
        return open(self.temp_filename_out, 'r',encoding='utf-8').read()

    def _run_non_interactive(self,api, text):
        assert api
        assert text
        self._write_to_temp_file(text)
        proc = subprocess.run(self.APIs_map[api]+['-i',self.temp_filename_in,'-o',self.temp_filename_out],capture_output=True)
        return True if proc.returncode == 0 else False
    
    def _run_interactive(self, text):
        assert text
        assert self.interactive
        assert self.task_proc
        text+='\n'
        self.task_proc.stdin.write(str.encode(text))
        self.task_proc.stdin.flush()
        return self.task_proc.stdout.readline().decode('utf8').strip()
    
    def _do_task(self, task=None, text=None):
        if self.interactive:
            if self.task==task:
                return self._run_interactive(text)
            else:
                raise Exception("You initialized the interactive object with {self.task} task and you are asking for another task. You may create another object with {task} task.")
        
        if self._run_non_interactive(api=task, text=text):
            return self._read_from_temp_file()
        raise Exception('Internal Error occured!')

    def segment(self, text):
        return self._do_task(task='segment',text=text)

    def stem(self, text):
        return self._do_task(task='stem',text=text)

    def diacratize(self):
        pass

    def POSify(self):
        pass

    def terminate_task(self):
        self.task_proc.terminate()


import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="farasapy",
    version="0.0.11",
    author="MagedSaeed",
    author_email="mageedsaeed1@gmail.com",
    description="A Python Wrapper for the well Farasa toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MagedSaeed/farasapy",
    packages=["farasa"],
    package_dir={"farasa": "farasa"},
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Topic :: Text Processing :: Linguistic",
        "Natural Language :: Arabic",
        "Intended Audience :: Science/Research",
        "Environment :: Console",
    ],
    python_requires=">=3.6",
    install_requires=["requests", "tqdm",],
)

# scripts=['farasa/__init__.py','farasa/farasa_bin/*','farasa/farasa_bin/lib/*','farasa/tmp/*'],



import io
import logging
import os
import re
import subprocess
import sys
import tempfile
import warnings
import zipfile
from pathlib import Path

import requests
from tqdm import tqdm


class FarasaBase:
    task = None
    __base_dir = Path(__file__).parent.absolute()
    __bin_dir = Path(f"{__base_dir}/farasa_bin")
    __bin_lib_dir = Path(f"{__bin_dir}/lib")

    # shlex not compatible with Windows replace it with list()
    # set java encoding with option `-Dfile.encoding=UTF-8`
    __BASE_CMD = ["java", "-Dfile.encoding=UTF-8", "-jar"]
    __APIs = {
        "segment": __BASE_CMD + [str(__bin_lib_dir / "FarasaSegmenterJar.jar")],
        "stem": __BASE_CMD
        + [str(__bin_lib_dir / "FarasaSegmenterJar.jar"), "-l", "true"],
        "NER": __BASE_CMD + [str(__bin_dir / "FarasaNERJar.jar")],
        "POS": __BASE_CMD + [str(__bin_dir / "FarasaPOSJar.jar")],
        "diacritize": __BASE_CMD + [str(__bin_dir / "FarasaDiacritizeJar.jar")],
    }
    __interactive = False
    __task_proc = None
    logger = None

    def __init__(self, interactive=False, logging_level="WARNING"):
        self._config_logs(logging_level)
        self.logger.debug("perform system check...")
        self.logger.debug("check java version...")
        self._check_java_version()
        self.logger.debug("check toolkit binaries...")
        self._check_toolkit_binaries()
        Path(f"{self.__base_dir}/tmp").mkdir(exist_ok=True)
        self.logger.info("Dependencies seem to be satisfied..")
        if interactive:
            self.__interactive = True
            self.logger.warning(
                "Be careful with large lines as they may break on interactive mode. You may switch to Standalone mode for such cases."
            )
            self.logger.info(
                f"\033[37minitializing [{self.task.upper()}] task in \033[32mINTERACTIVE \033[37mmode..."
            )
            self._initialize_task()
            self.logger.info(
                f"task [{self.task.upper()}] is initialized interactively."
            )
        else:
            self.logger.info(
                f"task [{self.task.upper()}] is initialized in \033[34mSTANDALONE \033[37mmode..."
            )

    def _config_logs(self, logging_level):
        self.logger = logging.getLogger("farasapy_logger")
        self.logger.propagate = False
        self.logger.setLevel(getattr(logging, logging_level.upper()))
        logs_formatter = logging.Formatter(
            "[%(asctime)s - %(name)s - %(levelname)s]: %(message)s"
        )
        if not self.logger.hasHandlers():
            stream_logger = logging.StreamHandler()
            stream_logger.setFormatter(logs_formatter)
            self.logger.addHandler(stream_logger)

    def _check_toolkit_binaries(self):
        download = False
        # check in bin folder:
        for jar in ("FarasaNERJar", "FarasaPOSJar", "FarasaDiacritizeJar"):
            if not Path(f"{self.__bin_dir}/{jar}.jar").is_file():
                download = True
                break

        if (
            download
            or not Path(f"{self.__bin_lib_dir}/FarasaSegmenterJar.jar").is_file()
        ):  # last check for binaries in farasa_bin/lib
            self.logger.info("some binaries are not existed.")
            self._download_binaries()

    def _get_content_with_progressbar(self, request):
        totalsize = int(request.headers.get("content-length", 0))
        blocksize = 3413334
        bar = tqdm(
            total=totalsize,
            unit="iB",
            unit_scale=True,
            ncols=5,
            dynamic_ncols=True,
            file=sys.stdout,
        )
        content = None
        for data in request.iter_content(blocksize):
            bar.update(len(data))
            if content is None:
                content = data
            else:
                content += data
        return content

    def _download_binaries(self):
        self.logger.info("downloading zipped binaries...")
        try:
            # change download url from github releases to qcri
            # binaries_url = "https://github.com/MagedSaeed/farasapy/releases/download/toolkit-bins-released/farasa_bin.zip"
            binaries_url = "https://farasa-api.qcri.org/farasapy/releases/download/toolkit-bins-released/farasa_bin.zip"
            binaries_request = requests.get(binaries_url, stream=True, verify=False)
            # show the progress bar while getting content
            content_bytes = self._get_content_with_progressbar(binaries_request)
            self.logger.debug("extracting...")
            binzip = zipfile.ZipFile(io.BytesIO(content_bytes))
            binzip.extractall(path=self.__base_dir)
            self.logger.debug("toolkit binaries are downloaded and extracted.")
        except Exception as e:
            self.logger.error("an error occured")
            self.logger.error(e)

    def __initialize_task_proc(self):
        self.__task_proc = subprocess.Popen(
            self.__APIs[self.task],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    def _initialize_task(self):
        word = "اختبار"
        word += "\n"
        bword = str.encode(word)
        self.__initialize_task_proc()
        return self._run_task_interactive(bword)

    def _check_java_version(self):
        try:
            version_proc_output = subprocess.check_output(
                ["java", "-version"], stderr=subprocess.STDOUT, encoding="utf8"
            )
            version_pattern = r"\"(\d+\.\d+).*\""
            java_version = float(
                re.search(version_pattern, version_proc_output).groups()[0]
            )
            if java_version >= 1.7:
                self.logger.debug(
                    f"Your java version is {java_version} which is compatiple with Farasa "
                )
            else:
                warnings.warn(
                    "You are using old version of java. Farasa is compatiple with Java 7 and above "
                )
        except subprocess.CalledProcessError as proc_err:
            self.logger.error(f"error occured: {proc_err}.")
            raise Exception(
                "We could not check for java version on the machine. Please make sure you have installed Java 1.7+ and add it to your PATH."
            )

    def _run_task(self, btext):
        assert btext is not None
        tmpdir = str(self.__base_dir / "tmp")
        # if delete=True on Windows cannot get any content
        # https://docs.python.org/3/library/tempfile.html#tempfile.NamedTemporaryFile
        itmp = tempfile.NamedTemporaryFile(dir=tmpdir, delete=False)
        otmp = tempfile.NamedTemporaryFile(dir=tmpdir, delete=False)
        try:
            itmp.write(btext)
            # https://stackoverflow.com/questions/46004774/python-namedtemporaryfile-appears-empty-even-after-data-is-written
            itmp.flush()
            proc = subprocess.run(
                self.__APIs[self.task] + ["-i", itmp.name, "-o", otmp.name],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                # this only compatiple with python>3.6
                # capture_output=True,
            )
            if proc.returncode == 0:
                result = otmp.read().decode("utf8").strip()
            else:
                self.logger.critical(
                    f"error occurred! stdout: , {proc.stdout},  stderr: , {proc.stderr}"
                )
                self.logger.critical(f"return code: {proc.returncode}")
                raise Exception("Internal Error occurred!")
        finally:
            itmp.close()
            otmp.close()
            os.unlink(itmp.name)
            os.unlink(otmp.name)
        return result

    def _run_task_interactive(self, btext):
        assert isinstance(btext, bytes)
        assert self.__interactive
        try:
            self.__task_proc.stdin.flush()
            self.__task_proc.stdin.write(btext)
            self.__task_proc.stdin.flush()
        except BrokenPipeError as broken_pipe:
            self.logger.error(
                f"pipe broke! error code and message: [{broken_pipe}]. reinitailize the process.., This may take sometime depending on the running task"
            )
            self.__initialize_task_proc()
            self.__task_proc.stdin.flush()
            self.__task_proc.stdin.write(btext)
            self.__task_proc.stdin.flush()

        output = self.__task_proc.stdout.readline().decode("utf8").strip()
        self.__task_proc.stdout.flush()
        return output

    def _do_task_interactive(self, strip_text):
        outputs = []
        for line in strip_text.split("\n"):
            newlined_line = line + "\n"
            byted_newlined_line = str.encode(newlined_line)
            output = self._run_task_interactive(byted_newlined_line)
            if output:
                outputs.append(output)
        return "\n".join(outputs)

    def _do_task_standalone(self, strip_text):
        byted_strip_text = str.encode(strip_text)
        return self._run_task(btext=byted_strip_text)

    def _do_task(self, text):
        strip_text = text.strip()
        if self.__interactive:
            return self._do_task_interactive(strip_text)
        else:
            return self._do_task_standalone(strip_text)

    def terminate(self):
        self.__task_proc.terminate()
